#!/usr/bin/python3
# -*- coding: utf-8 -*-
__author__ = "legendzdy@dingtalk.com"
"""
Author: legendzdy@dingtalk.com
Data: 20250210
Description:
This Script of raw fq data from Legendzdy.  Of course, this means there's a possibility
for other ways.Use at your own discretion.
"""
import argparse
import re
import gzip
import pysam
import concurrent.futures

class getModifications:
    BASE_PATTERNS = {
        'A': re.compile(r'A'),
        'T': re.compile(r'T'),
        'C': re.compile(r'C')
    }

    def __init__(self, inputFq, samFile, bedFile, output, workers=4):
        self.inputFq = inputFq
        self.samFile = samFile
        self.bedFile = bedFile
        self.output = output
        self.workers = workers

    def getAnnotation(self):
        def read_bed_file():
            with open(self.bedFile, 'r') as f:
                for line in f:
                    line = line.strip().split("\t")
                    gid = line[0] + ":" + line[1]
                    enst = line[3]
                    base = line[4]
                    yield gid, (enst, base)
        return dict(read_bed_file())

    def posFind(self, modList):
        counter = 0
        posList = []
        for mod in modList:
            counter += int(mod)
            posList.append(counter)
            counter += 1
        return posList

    # 找到修饰位置和对应的p值
    def intersectionPosition(self, mm, baseList, mlNum):
        modList = mm.split(",")[1:]
        posList = self.posFind(modList)
        mlcount = 0
        readPos = []
        readPosPvalue = []
        for pos in posList:
            rpos = baseList[pos]
            pvalue = round(int(mlNum[mlcount])/255, 3)
            readPos.append(rpos)
            readPosPvalue.append(pvalue)
            mlcount += 1
        return readPos, readPosPvalue


    def process_record(self, name, seq):
        abase = re.compile(r'A')
        tbase = re.compile(r'T')
        cbase = re.compile(r'C')
        namelist = name.split("\t")
        id = namelist[0].split("@")[-1]  # 获取ID
        MMtagList = namelist[-2].split("MM:Z:")[-1].split(";")
        MLtag = namelist[-1].split(",")[1:]
        Alist = [match.start() + 1 for match in abase.finditer(seq)]
        Tlist = [match.start() + 1 for match in tbase.finditer(seq)]
        Clist = [match.start() + 1 for match in cbase.finditer(seq)]
        mlstart = 0
        rpos_mod = {}
        for mm in MMtagList:
            mmlen = len(mm.split(","))
            if mmlen > 1:
                mlend = mlstart + mmlen - 1
                mlNum = MLtag[mlstart:mlend]
                mlstart += mmlen - 1
                mod_map = {
                    'A+a': Alist,
                    'A+17596': Alist,
                    'T+17802': Tlist,
                    'C+m': Clist
                }
                for mod, base_list in mod_map.items():
                    if mm.startswith(mod):
                        readPos, readPosPvalue = self.intersectionPosition(mm, base_list, mlNum)
                        for rpos, pvalue in zip(readPos, readPosPvalue):
                            if pvalue > 0.98:
                                if rpos_mod.get(id) is None:
                                    rpos_mod[id] = [(rpos, pvalue, mod)]
                                else:
                                    rpos_mod[id].append((rpos, pvalue, mod))
        return rpos_mod

    def get_mod_position_with_read(self):
        rpos_mod = {}
        with gzip.open(self.inputFq, 'rt') as f:
            with concurrent.futures.ThreadPoolExecutor(max_workers=self.workers) as executor:
                futures = []
                while True:
                    try:
                        # 读取四行fastq格式数据
                        name = next(f).strip()
                        seq = next(f).strip()
                        plus = next(f).strip()
                        qual = next(f).strip()
                        futures.append(executor.submit(self.process_record, name, seq))
                    except StopIteration:
                        break

                for future in concurrent.futures.as_completed(futures):
                    result = future.result()
                    for id, mod_info in result.items():
                        if id in rpos_mod:
                            rpos_mod[id].extend(mod_info)
                        else:
                            rpos_mod[id] = mod_info

        return rpos_mod


    def get_mod_position_with_sam(self):
        rpos_mod = self.get_mod_position_with_read()
        annot = self.getAnnotation()
        with open(self.output, "w") as f:
            for read in pysam.AlignmentFile(self.samFile, "r"):
                if not read.is_unmapped and rpos_mod.get(read.query_name) is not None:
                    pos = read.reference_start
                    read_pos = 1
                    strand = "-" if read.is_reverse else "+"
                    # 循环遍历 cigartuples，将读取的读和基因组位置合并后直接写入文件
                    for op, length in read.cigartuples:
                        if op == 0:  # M: 匹配或错配
                            # 匹配的碱基在基因组上的位置
                            for i in range(length):
                                mod_list = rpos_mod[read.query_name]
                                for rpos, pvalue, mod in mod_list:
                                    if rpos == read_pos + i:
                                        genome_id = f"chr{read.reference_name}:{pos + i}"
                                        if annot.get(genome_id) is not None:
                                            enst, base = annot[genome_id]
                                            f.write(f"{read.reference_name}\t{pos + i}\t{pos + i+1}\t{pvalue}\t{enst}\t{strand}\t{mod}\t{base}\n")
                            pos += length
                            read_pos += length
                        elif op == 1:  # I: 插入
                            # 插入的碱基在参考基因组上不占位置，但在 read 上有位置
                            read_pos += length
                        elif op == 2:  # D: 删除
                            # 删除的碱基在参考基因组上仍占位置，但跳过它们
                            pos += length
                        elif op == 4:  # S: 跳过
                            # 跳过的碱基不占基因组位置，但在 read 上有位置
                            read_pos += length
                        elif op == 5:  # H: 硬剪切
                            # 硬剪切不占基因组位置
                            read_pos += length
                        else:
                            print(f"op: {op}, length: {length} not supported")

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Parallel modifacts detection from raw fastq data.'
    )
    parser.add_argument('--version', action='version', version='modkit 1.0')
    parser.add_argument('-i', '--inputFq', required=True)
    parser.add_argument('-s', '--samFile', required=True)
    parser.add_argument('-b', '--bedFile', required=True)
    parser.add_argument('-o', '--output', required=True)
    parser.add_argument('-w', '--workers', default=4, type=int)
    args = parser.parse_args()
    mod = getModifications(args.inputFq, args.samFile, args.bedFile, args.output, args.workers)
    mod.get_mod_position_with_sam()



#!/usr/bin/env python3
__author__ = "legendzdy@dingtalk.com"
"""
Author: legendzdy@dingtalk.com
Data: 20200603
Description:
This Script of ONTrnaDirectSeq from Legendzdy.  Of course, this means there's a possibility
for other ways.Use at your own discretion.
"""
import argparse
import re


def random_forest_data(samFile, base, fout):
    """
    This function will organize a dict of features.
    :param samFile: Pandas DataFrame to subset the two classes from
    :param fout: outfile
    """
    transcript_lengths = {}
    S = re.compile(r'\d+S')
    if base == "U":
        T = re.compile(r'T|t')
    elif base == "A":
        T = re.compile(r'A|a')
    elif base == "G":
        T = re.compile(r'G|g')
    elif base == "C":
        T = re.compile(r'C|c')
    else:
        print("Plase choose one of ATGC ")
    M = re.compile(r'\d+M')
    cstr = re.compile(r'cs:Z\S+')
    identical_block = re.compile(r':[0-9]+')
    deltion_block = re.compile(r'[\-][A-Za-z]+')
    insertion_block = re.compile(r'[\+][A-Za-z]+')
    substituted_ref_read = re.compile(r'\*[a-z][a-z]')

    with open(fout, "w") as fo:
        with open(samFile, "r") as fi:
            if base == "U":
                fo.write(
                    "read,UmisRate,UmisQrate,t_a_rate,t_a_Qrate,t_c_rate,t_c_Qrate,t_g_rate,t_g_Qrate,del_rate\n")
            elif base == "A":
                fo.write(
                    "read,AmisRate,AmisQrate,a_t_rate,a_t_Qrate,a_c_rate,a_c_Qrate,a_g_rate,a_g_Qrate,del_rate\n")
            elif base == "G":
                fo.write(
                    "read,GmisRate,GmisQrate,g_a_rate,g_a_Qrate,g_c_rate,g_c_Qrate,g_t_rate,g_t_Qrate,del_rate\n")
            elif base == "C":
                fo.write(
                    "read,CmisRate,CmisQrate,c_a_rate,c_a_Qrate,c_t_rate,c_t_Qrate,c_g_rate,c_g_Qrate,del_rate\n")
            else:
                print("Plase choose one of AUGC ")
            while True:
                try:
                    line = next(fi).strip()
                    if line.startswith('@'):
                        if line.startswith('@SQ'):
                            name = line[line.find('SN:') + 3:].split()[0]
                            length = float(
                                line[line.find('LN:') + 3:].split()[0])
                            transcript_lengths[name] = length
                        continue

                    if line.split()[1] == "0" or line.split()[1] == "16":
                        cs = line[cstr.search(line).span()[
                            0] + 5:cstr.search(line).span()[1]]
                        read, transcript, cigar, quality, sequence, Q = line.split()[0], line.split(
                        )[2], line.split()[5], int(line.split()[4]), line.split()[9], line.split()[10]
                        matchLen = sum([int(i[:-1]) for i in M.findall(cigar)])
                        coverage = matchLen/transcript_lengths[transcript]
                        if coverage < 0.8:
                            continue

                        readStart, UmisNum, UmisQ, delNum, t_a, t_c, t_g, t_aQ, t_cQ, t_gQ, matchseq = 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, ""
                        if S.match(cigar) != None:
                            numS = int(cigar[S.match(cigar).span()[
                                       0]:S.match(cigar).span()[1] - 1])
                            readStart = numS
                        while len(cs) != 0:
                            if identical_block.match(cs) != None:
                                idenum = int(cs[identical_block.match(cs).span()[
                                             0]+1:identical_block.match(cs).span()[1]])
                                if len(matchseq) == 0:
                                    matchseq += str(sequence[readStart:idenum])
                                else:
                                    matchseq += str(
                                        sequence[readStart:idenum+readStart])
                                readStart += idenum
                                cs = cs[identical_block.match(cs).span()[1]:]
                            elif deltion_block.match(cs) != None:
                                delStr = cs[deltion_block.match(cs).span(
                                )[0]+1:deltion_block.match(cs).span()[1]]
                                cs = cs[deltion_block.match(cs).span()[1]:]
                                if len(T.findall(delStr)) != 0:
                                    delNum += len(T.findall(delStr))
                            elif insertion_block.match(cs) != None:
                                readStart += len(cs[insertion_block.match(cs).span()[
                                                 0]+1:insertion_block.match(cs).span()[1]])
                                cs = cs[insertion_block.match(cs).span()[1]:]
                            elif substituted_ref_read.match(cs) != None:
                                refStr = cs[substituted_ref_read.match(cs).span(
                                )[0]+1:substituted_ref_read.match(cs).span()[1]][0]
                                readStr = cs[substituted_ref_read.match(cs).span(
                                )[0]+1:substituted_ref_read.match(cs).span()[1]][1]
                                cs = cs[substituted_ref_read.match(cs).span()[
                                    1]:]
                                if refStr == "t" and base == "U":
                                    accuracy = 1 - \
                                        0.1 ** ((ord(Q[readStart-1]) - 33) / 10)
                                    UmisNum += 1
                                    UmisQ += accuracy
                                    if readStr == "a":
                                        t_a += 1
                                        t_aQ += accuracy
                                    elif readStr == "c":
                                        t_c += 1
                                        t_cQ += accuracy
                                    else:
                                        t_g += 1
                                        t_gQ += accuracy
                                elif refStr == "a" and base == "A":
                                    accuracy = 1 - \
                                        0.1 ** ((ord(Q[readStart - 1]) - 33) / 10)
                                    UmisNum += 1
                                    UmisQ += accuracy
                                    if readStr == "t":
                                        t_a += 1
                                        t_aQ += accuracy
                                    elif readStr == "c":
                                        t_c += 1
                                        t_cQ += accuracy
                                    else:
                                        t_g += 1
                                        t_gQ += accuracy
                                elif refStr == "g" and base == "G":
                                    accuracy = 1 - \
                                        0.1 ** ((ord(Q[readStart - 1]) - 33) / 10)
                                    UmisNum += 1
                                    UmisQ += accuracy
                                    if readStr == "a":
                                        t_a += 1
                                        t_aQ += accuracy
                                    elif readStr == "c":
                                        t_c += 1
                                        t_cQ += accuracy
                                    else:
                                        t_g += 1
                                        t_gQ += accuracy
                                elif refStr == "c" and base == "C":
                                    accuracy = 1 - \
                                        0.1 ** ((ord(Q[readStart - 1]) - 33) / 10)
                                    UmisNum += 1
                                    UmisQ += accuracy
                                    if readStr == "a":
                                        t_a += 1
                                        t_aQ += accuracy
                                    elif readStr == "t":
                                        t_c += 1
                                        t_cQ += accuracy
                                    else:
                                        t_g += 1
                                        t_gQ += accuracy
                            else:
                                print("error in cs")
                                break
                        UmatchNum = len(T.findall(matchseq))
                        read_mean_Q = sum(
                            [1 - 0.1 ** ((ord(Q[i]) - 33) / 10) for i in range(len(Q))])/len(Q)
                        if UmisNum != 0:
                            UmisRate = UmisNum/(delNum+UmisNum+UmatchNum)
                            UmisQrate = (UmisQ/UmisNum)*read_mean_Q
                            t_a_rate = t_a / (delNum+UmisNum+UmatchNum)
                            t_c_rate = t_c / (delNum+UmisNum+UmatchNum)
                            t_g_rate = t_g / (delNum+UmisNum+UmatchNum)
                            if t_a != 0:
                                t_a_Qrate = (t_aQ/t_a)*read_mean_Q
                            else:
                                t_a_Qrate = 0
                            if t_c != 0:
                                t_c_Qrate = (t_cQ/t_c)*read_mean_Q
                            else:
                                t_c_Qrate = 0
                            if t_g != 0:
                                t_g_Qrate = (t_gQ/t_g)*read_mean_Q
                            else:
                                t_g_Qrate = 0
                        else:
                            UmisRate = 0
                            UmisQrate = 0
                            t_a_rate = 0
                            t_c_rate = 0
                            t_g_rate = 0
                            t_a_Qrate = 0
                            t_c_Qrate = 0
                            t_g_Qrate = 0
                        if delNum != 0:
                            del_rate = delNum/(delNum+UmisNum+UmatchNum)
                        else:
                            del_rate = 0

                        fo.write("{},{:.4f},{:.4f},{:.4f},{:.4f},{:.4f},{:.4f},{:.4f},{:.4f},{:.4f}\n".format(
                            read, UmisRate, UmisQrate, t_a_rate, t_a_Qrate, t_c_rate, t_c_Qrate, t_g_rate, t_g_Qrate, del_rate))
                except StopIteration:
                    break


def main():
    parser = argparse.ArgumentParser(
        description='Extract length and quality information of sequences form fastq file')
    parser.add_argument('-s', '--input_sam', type=str, required=True,
                        help='sam file from minimap')
    parser.add_argument('-b', '--base', type=str, required=True,
                        help='choose one of AUGC')
    parser.add_argument('-o', '--out_file', type=str, required=False,
                        help='Path to an output file to be created,default= ./')
    args = parser.parse_args()
    random_forest_data(args.input_sam, args.base, args.out_file)


if __name__ == '__main__':
    main()
# random_forest_data("CTL2-0601.fastq.sam","test.csv")



import time
import typer
from typing_extensions import Annotated
from rich.progress import Progress, SpinnerColumn, TextColumn
from basebio import check_path_exists
from ..utils.sam_fatures_catch_all import random_forest_data
from ..utils.nascentRNA_fetch import new_fq
from ..utils.quantify import salmon_map, salmon_quantify

app = typer.Typer()

@app.command()
def nascentRNA(
    input: Annotated[str, typer.Option("--input", "-i", help="Input fastq files.")],
    sam: Annotated[str, typer.Option("--sam", "-s", help="mapping sam/bam file.")],
    reference: Annotated[str, typer.Option("--reference", "-r", help="reference transcripts path.")],
    base: Annotated[str, typer.Option("--base", "-b", help="choose one of AUGC.")],
    model: Annotated[str, typer.Option("--model", "-m", help="train model.")],
    output: Annotated[str, typer.Option("--output", "-o", help="Output file path.")]=".",
    prefix: Annotated[str, typer.Option("--prefix", "-p", help="Prefix for output files.")]="prefix",
    ):
    """
    Detect nascent RNA using mapping sam/bam file and fastq files. And quantify them using salmon.
    """
    
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        progress.add_task(description="Detecting nascent RNA start...", total=None)
        start=time.time()
        
        output_file=f"{output}/{prefix}_{base}.csv"
        progress.add_task(description="Getting fatures from sam files...", total=None)
        if not check_path_exists(output_file):
            random_forest_data(sam, base, output_file)
        progress.add_task(description=f"Getting fatures from sam files Done", total=None)

        output_fastq=f"{output}/{prefix}_nascentRNA.fastq"
        progress.add_task(description="Getting nascentRNA reads...", total=None)
        if not check_path_exists(output_fastq):
            new_fq(output_file, model, input, output_fastq)
        progress.add_task(description="Getting nascentRNA reads Done", total=None)

        output_bam=f"{output}/{prefix}_nascentRNA.bam"
        progress.add_task(description="Mapping nascentRNA reads to reference...", total=None)
        if not check_path_exists(output_bam):
            salmon_map(output_fastq, reference, output_bam)
        progress.add_task(description="Mapping nascentRNA reads to reference Done", total=None)

        output_quant=f"{output}/{prefix}_quant"
        progress.add_task(description="Quantifying nascentRNA reads ...", total=None)
        if not check_path_exists(output_quant):
            salmon_quantify(output_bam, reference, output_quant)
        progress.add_task(description="Quantifying nascentRNA reads Done", total=None)

        end=time.time()
        time_cost=f"{(end - start) // 3600}h{((end - start) % 3600) // 60}m{(end - start) % 60:.2f}s"
        print(f"Detect nascent RNA Done, time cost: {time_cost}")
        progress.add_task(description=f"Detect nascent RNA Done, time cost: {time_cost}", total=None)


# -*- coding: utf-8 -*-
__author__ = "legendzdy@dingtalk.com"
"""
Author: legendzdy@dingtalk.com
Data: 20250612
Description:
function map.
"""
from basebio import run_command

def minimap2map(input, reference, output, tool, params, threads):
    """
    map with minimap2.
    Args:
        input: input fasta file.
        reference: reference fasta file.
        output: output sam/bam file.
        threads: number of threads.
        tool: minimap2 tool path.
        params: minimap2 params.
    """
    params_list = params.split()
    suffix = output.split(".")[-1]
    if suffix == "sam":
        command = [
            tool, *params_list, "-t", str(threads),
            reference, input, "-o", output
            ]
        use_shell = False
    elif suffix == "bam":
        command = [
            tool, *params_list, "-t", str(threads),
            reference, input, "|", "samtools", 
            "sort", "-@", str(threads), "-o", output, 
            "&&", "samtools", "index", output
            ]
        use_shell = True
    
    run_command(command, use_shell=use_shell)


# -*- coding: utf-8 -*-
__author__ = "legendzdy@dingtalk.com"
"""
Author: legendzdy@dingtalk.com
Data: 20250612
Description:
function map.
"""
from basebio import run_command
from .map import minimap2map

def salmon_map(input, reference, output):
    """
    Map with minimap2.
    Args:
        input: input fasta file.
        reference: reference transcripts fasta file.
        output: output bam file.
    """
    tool = "minimap2"
    parms = "--secondary=no --cs -a"
    threads = 4
    minimap2map(input, reference, output, tool, parms, threads)
    

def salmon_quantify(input, reference, output):
    """
    map with minimap2.
    Args:
        input: input bam file.
        reference: reference transcripts fasta file.
        output: output quantification file.
    """
    
    command = [
        "salmon", "quant", "-p", "4",
        "-t", reference, "-l", "SF", 
        "-a", input, "-o", output
        ]
    run_command(command)